#include <cassert>
#include "utils.h"
#include "AST.h"
#include "CodeContext.h"
#include "CfSource.h"
#include "clone.h"
#include "code.h"
#include "code-util.h"
#include "code-grid.h"
#include "garbageList.h"
#include "nullify-local.h"
#include "optimize.h"
#include "lower.h"
#include "version.h"
#include "code-MIVE.h"
#include "Worklist.h"
#include "ullist.h"
#include "cfg.h"

extern int compile_verbose_level;
extern void SRsetup();

static void staticAnal(TreeNode *method, bool dominators,
		       bool ignoreNSE = false, bool computeDeps = false);

static TreeNode *optimize(TreeNode *t);

void DummyNode::emitStatement( CodeContext &context ) {}

PrimitiveLitNode *newBoolLit(bool b, SourcePosn p)
{
  return new PrimitiveLitNode(Literal(b), p);
}

TreeNode *CloneTree(TreeNode *t, bool cleanUseDef)
{
  TreeNode *newNode = t->clone();
  if (cleanUseDef) {
    newNode->setUses(NULL);
    newNode->setDefs(NULL);
  }
  const int childCount = newNode->arity();
  for (int sweep = 0; sweep < childCount; ++sweep)
    newNode->child(sweep, CloneTree(newNode->child(sweep), cleanUseDef));

  return newNode;
}

static bool assignablenode(const TreeNode *t) {
  return t->hasLval();
}

static bool objectnode(const TreeNode *t)
{
  return isObjectNode(t);
}

static bool exprnode(const TreeNode *t)
{
  return isExprNode(t);
}

/////////////////////////////////////////////////////////////////////////////
// Foreach Opts 
/////////////////////////////////////////////////////////////////////////////

/* The enclosing loop of a foreach loop, if any */
static map_tree_to_tree enclLoop;

#define REPORT_AND_ADD(x, o)			\
((DEBUG_OPT ?					\
  ((o << "Add "),				\
   (o << PLCODE(x)),				\
   (o << " to worklist"),			\
   (o << endl), 0) :				\
  0),						\
(w)->add((x)))

#define ADD(x, w) do if (isExprNode((x))) REPORT_AND_ADD((x), cout); while (0)

void propagateChanges(Worklist *w, TreeNode *t)
{
  ADD(t->parent(), w);
  foreach (u, llist<TreeNode *>, *(t->getUses()))
    ADD(*u, w);
}

#undef ADD
#undef REPORT_AND_ADD

/* PROPAGATE(w) is used as (lambda (t) (propagateChanges w t)), but since
   we don't have lambda we use current_worklist as a hackish way to store w. */

Worklist *current_worklist;

#if 0
/* unused */
static void propToCurrentWorklist(TreeNode *t)
{
  propagateChanges(current_worklist, t);
}

#define PROPAGATE(w) \
    ((current_worklist = (w)), propToCurrentWorklist)
#endif /* 0 */

static void invAndPropToCurrentWorklist(TreeNode *t, TreeNode *WRTloop)
{
  if (!isInvar(t, WRTloop)) {
    foundInvariant(t, WRTloop);
    propagateChanges(current_worklist, t);
  }
}

/* similar to PROPAGATE(w), above */
#define INVARIANT_AND_PROPAGATE(w) \
    ((current_worklist = (w)), invAndPropToCurrentWorklist)

extern void gotMIVE(TreeNode *t, TreeNode *WRTloop, MIVE *m);

#define GOT_MIVE(w) \
    ((current_worklist = (w)), gotMIVE)

/////////////////////////////////////////////////////////////////////////////
// Figure out if t is invariant and/or if we can put a polynomial on it.
/////////////////////////////////////////////////////////////////////////////

static void computeMIVE(TreeNode *t, TreeNode *WRTloop, treeSet *loopContents,
			Worklist *w) 
{
  /* If t is the child of an assignment, work on t's parent instead. */
  if (isAssignNode(t->parent())) {
    w->add(t->parent());
    return;
  }

  /* If t is already known to be invariant there's no need to
     recompute anything, unless t is an assignment, in which case we
     may have improved MIVE info to deal with. */
  bool recompute_all = !isInvar(t, WRTloop), assign = isAssignNode(t);

  /* Determine and record whether it is an invariant. */
  if (recompute_all) 
    t->loopInvariant(WRTloop, loopContents, INVARIANT_AND_PROPAGATE(w));

  /* Try to compute a polynomial describing it. */
  if (recompute_all || assign)
    t->computeMIVE(WRTloop, GOT_MIVE(w));
}

/* Build a list that contains the given set of nodes. */
llist<TreeNode *> * set_to_list(const treeSet *s)
{
  llist<TreeNode *> * result = NULL;
  for (treeSet::const_iterator i = s->begin(); i != s->end(); ++i)
    result = cons((TreeNode *) *i, result);
  return result;
}

/* Build a treeSet that contains the given list of nodes. */
treeSet * list_to_set(llist<TreeNode *> *l)
{
  return new treeSet(l->begin(), l->end());
}

CFGset * list_to_CFGset(llist<CfgNode *> *l)
{
  return new CFGset(l->begin(), l->end());
}

/* The loopAnal pass tags each foreach loop with a list of loop invariants
  and computes polynomials (over invariants and induction variables) that
  represent what certain exprs are computing.  The workhorse is foreachAnal().
*/

void TreeNode::loopAnal(TreeNode *enclosingLoop)
{
  for (int i = arity(); i-- > 0; )
    child(i)->loopAnal(enclosingLoop);
}

#define pf(m, i) (m)->p[(i)]->asPolyFactor()

/* This will terminate because at each step either the worklist gets
   smaller or the information we have gets better.  (And the info
   can only get better a finite number of times.)
*/
static void foreachAnal(ForEachStmtNode *l)
{
  if (DEBUG_LOOP_ANAL) {
    cout << endl << "foreachAnal: " << l->position().asString()
	 << endl << endl;
  }
  Worklist w(l);
  treeSet *loopContents = list_to_set(cons(l->vars()->child(0), w.asList()));
  w.filter(exprnode);
  MIVEcontext *e = MIVEcontextOfLoop(l) = new MIVEcontext();
  TreeNode *pairnode = l->vars()->child(0);
  gotMIVE(pairnode->simpName(), l,
	  e->adjoin(pairnode->simpName()->ident(),
		    pairnode->initExpr()->type()->tiArity(),
		    l));

  while (!w.empty()) {
    TreeNode *t = w.pop();
    if (isExprNode(t)) {
      // if (isAssignNode(t->parent())) continue;

      if (DEBUG_LOOP_ANAL) {
	cout << endl << "Working on: "; cout << PCODE(t); cout << endl << endl;
      }
      computeMIVE(t, l, loopContents, &w);

    } else {
      // should not happen
      cout << "Not an ExprNode: ";
      t->pseudoprint(cout, 2);
      cout << endl;
    }
  }
  delete loopContents;
  
  setDerivs(l);
}

void ForEachStmtNode::loopAnal(TreeNode *enclosingLoop)
{
  if (compile_verbose_level >= 3)
    cout << "   loop analysis of foreach at " << position().asString() << endl;
  enclLoop[this] = enclosingLoop;
  initInvar(this);

  for (int i = arity(); i-- > 0; )
    child(i)->loopAnal(this);

  foreachAnal(this);
}

/////////////////////////////////////////////////////////////////////////////
// utils
/////////////////////////////////////////////////////////////////////////////

llist<TreeNode *> *& ForEachStmtNode::junk_post() { return _junk_post; }
llist<TreeNode *> *& ForEachStmtNode::junk_pre() { return _junk_pre; }

/* Remove all statements in s from t. */
TreeNode *deleteStmts(TreeNode *t, set<StatementNode *> *s)
{
  treeSet ss;
  for (set<StatementNode *>::iterator i = s->begin(); i != s->end(); i++)
    ss.insert(*i);
  return t->deleteStmts(&ss);
}


// deleteStmts() behavior is documented in AST.h

TreeNode *TreeNode::deleteStmts(treeSet *s)
{
  const int childCount = arity();
  for (int sweep = 0; sweep < childCount; ++sweep)
    child(sweep, child(sweep)->deleteStmts(s));

  return this;
}

TreeNode *BlockNode::deleteStmts(treeSet *s)
{
  TreeNode *l = child(0);
  TreeNode *result = this;
  int c = l->arity();
  bool del = false; // whether any top-level stmts in the block must be deleted

#if 0
  cout << endl << "Deleting stmts from block:" << endl;
  pseudoprint(cout, 0);
  cout << endl;
#endif

  for (int i = 0; i < c; ++i)
    if (contains(s, l->child(i)) ||
	isExpressionStmtNode(l->child(i)) &&
	contains(s, l->child(i)->expr()))
      del = true;
    else
      l->child(i, l->child(i)->deleteStmts(s));
  
  // If top-level stmts must be deleted then rebuild the block from scratch.
  if (del) {
    llist<TreeNode *> *q = NULL;
    while (c-- > 0)
      if (!(contains(s, l->child(c)) ||
	    isExpressionStmtNode(l->child(c)) &&
	    contains(s, l->child(c)->expr())))
	q = cons(l->child(c), q);
      else {
        //l->child(c)->free();
      }
    //result = q ? (TreeNode *)new BlockNode(q, position()) : (TreeNode *)new EmptyStmtNode();
    stmts(new TreeListNode(q, position())); // DOB: prevent regenerating the BlockNode 
    //delete l;
  }

#if 0
  cout << endl << "Result of deleting stmts from block:" << endl;
  result->pseudoprint(cout, 0);
  cout << endl;
#endif

  return result;
}

TreeNode *StatementNode::deleteStmts(treeSet *s)
{
  if (contains(s, this)) {
    //this->free();
    return new EmptyStmtNode();
  }
  else
    return TreeNode::deleteStmts(s);
}

TreeNode *ExpressionStmtNode::deleteStmts(treeSet *s)
{
  if (contains(s, expr())) {
    //this->free();
    return new EmptyStmtNode();
  }
  else
    return StatementNode::deleteStmts(s);
}

/* Delete decls in s from t and return t or a rewritten version of t. */
TreeNode *deleteDeclsFromBlocks(TreeNode *t, treeSet *s)
{
  // This is safe, but it does rely on undocumented behavior of
  // deleteStmts(), above.
  return t->deleteStmts(s);
}

/////////////////////////////////////////////////////////////////////////////

static llist<CfgNode *> * current_list;
static void addToList(CfgNode *x)
{
  current_list = cons(x, current_list);
}
#define ADD_TO_LIST() ((current_list ?					\
			(free_all(current_list), current_list = NULL) :	\
			NULL),						\
		       addToList)

bool isFullDomainLoop(TreeNode *t)
{
  assert(isForEachStmtNode(t));
  llist<CfgNode *> *l;
  TraverseCfgScope(t, ADD_TO_LIST());
  l = dreverse(current_list);
  const CFGset *ls = list_to_CFGset(l);
  bool ignored = false;
  foreach (n, llist<CfgNode *>, *l) {
    (*n)->predIter(true, ls, &ignored);
    if (ignored) {
      cout << "isFullDomainLoop(): false at " << t->position().asString() <<
	endl;
      return false;
    }
  }
  return true;
}

#undef ADD_TO_LIST

/* Destructively replace all occurences of x in t with y. */
TreeNode *replaceNode(TreeNode *t, TreeNode *x, TreeNode *y)
{
  if (t == x)
    return y;
  else {
    int a = t->arity();
    for (int i = 0; i < a; i++)
      t->child(i, replaceNode(t->child(i), x, y));
    return t;
  }
}

/* Destructively replace any one occurence of x in t with y.  Returns t.
   Does assert(0) if x does not occur. */
TreeNode *replaceOneNode(TreeNode *t, TreeNode *x, TreeNode *y)
{
  if (t == x)
    return y;
  else {
    llist<TreeNode *> *todo = cons(t);
    while (todo != NULL) {
      TreeNode *c = todo->front();
      todo = todo->free();
      int a = c->arity();
      for (int i = 0; i < a; i++)
	if (c->child(i) == x) {
	  c->child(i, y);
	  free_all(todo);
	  return t;
	} else
	  push(todo, c->child(i));
    }
    assert(0);
    return t;
  }
}

/////////////////////////////////////////////////////////////////////////////

treeSet *allDefs(treeSet *vars)
{
  treeSet *result = new treeSet();
  for (treeSet::iterator i = vars->begin(); i != vars->end(); ++i) {
    treeSet *s = list_to_set(((TreeNode *) (*i))->getDefs());
    destructively_union(result, s);
    delete s;
  }
  return result;
}

treeSet *allUses(TreeNode *t)
{
  return list_to_set(t->getUses());
}

static inline bool opnd0_of_assignnode(TreeNode *t)
{
  TreeNode *p = t->parent();
  return (isAssignNode(p) && p->opnd0() == t);
}

static bool lval_context(TreeNode *t)
{
  while (isIIFAN(t->parent()))
    t = t->parent();
  return opnd0_of_assignnode(t);
}

// Return a set of uses that contains all Lvalue uses.
treeSet *LvalueUses(TreeNode *t)
{
  llist< TreeNode * > * const uses = t->getUses();
  treeSet *result = new treeSet;

  remove_copy_if( uses->begin(), uses->end(),
		  inserter( *result, result->begin() ),
		  not1( ptr_fun( lval_context ) ) );
  return result;
}

// Return a set of uses that contains all Rvalue uses.
treeSet *RvalueUses(TreeNode *t)
{
  llist< TreeNode * > * const uses = t->getUses();
  treeSet *result = new treeSet;

  remove_copy_if( uses->begin(), uses->end(),
		  inserter( *result, result->begin() ),
		  lval_context );
  return result;
}

treeSet *ObjectNodes(TreeNode *e)
{
  Worklist w(e);
  return list_to_set(w.filter(objectnode)->asList());
}

treeSet *AssignableNodes(TreeNode *e)
{
  Worklist w(e);
  return list_to_set(w.filter(assignablenode)->asList());
}


/////////////////////////////////////////////////////////////////////////////
// treeSet / CFGset utils
/////////////////////////////////////////////////////////////////////////////

/* True iff s is a set containing just e. */
bool isSingletonContaining(TreeNode *e, treeSet *s)
{
  return s->size() == 1 && *(s->begin()) == e;
}

treeSet *exprset_to_stmtset(treeSet *t)
{
  treeSet * result = new treeSet;
  for (treeSet::iterator i = t->begin(); i != t->end(); ++i)
    result = adjoin(result, enclosingStmt(*i));
  return result;
}

treeSet *convert_cfguset_to_treeset(const CFGuset *s)
{
  treeSet *result = new treeSet;
  for (CFGuset::const_iterator i = s->begin(); i != s->end(); ++i)
    adjoin(result, (*i)->astNode());
  return result;
}

map_tree_to_treeSet *
destructive_convert_cfgmap_to_astmap(map_CFGnode_to_CFGuset *r)
{
  map_tree_to_treeSet *result = new map_tree_to_treeSet;
  for (map_CFGnode_to_CFGuset::iterator i = r->begin(); i != r->end(); ++i)
    (*result)[(*i).first->astNode()] =
      convert_cfguset_to_treeset((*i).second);
  delete r;
  return result;
}

treeSet *destructive_convert_cfgset_to_treeset(CFGset *s)
{
  treeSet *result = new treeSet;
  for (CFGset::iterator i = s->begin(); i != s->end(); ++i)
    adjoin(result, (*i)->astNode());
  delete s;
  return result;
}

map_tree_to_treeSet *
destructive_convert_cfgmap_to_astmap(map_CFGnode_to_CFGset *r)
{
  map_tree_to_treeSet *result = new map_tree_to_treeSet;
  for (map_CFGnode_to_CFGset::iterator i = r->begin(); i != r->end(); ++i)
    (*result)[(*i).first->astNode()] =
      destructive_convert_cfgset_to_treeset((*i).second);
  delete r;
  return result;
}

// Add t to s if it isn't already there.
treeSet *treeSetAdjoin(treeSet *s, const TreeNode *t)
{
  return adjoin(s, t);
}

// Add t to s if it isn't already there.
treeSet *adjoin(treeSet *s, const TreeNode *t)
{
  s->insert(t);
  return s;
}

// Add t to s if it isn't already there.
CFGset *adjoin(CFGset *s, const CfgNode *t)
{
  s->insert((CfgNode *) t);
  return s;
}

bool contains(const CFGset *s, const CfgNode *t)
{
  return s->find((CfgNode *) t) != s->end();
}


bool contains(const treeSet *s, const TreeNode *t)
{
  return treeSetContains(s, t);
}

bool contains(const treeSet &s, const TreeNode *t)
{
  return treeSetContains(s, t);
}

bool treeSetContains(const treeSet *s, const TreeNode *t)
{
  return s != NULL && treeSetContains(*s, t);
}

bool treeSetContains(const treeSet &s, const TreeNode *t)
{
  return s.find(t) != s.end();
}

treeSet *copy(treeSet *s)
{
  return new treeSet(*s);
}

// Change s to s intersect t.
treeSet *destructively_intersect(treeSet *s, treeSet *t)
{
  treeSet::iterator i = s->begin();
  treeSet::iterator j = t->begin();
  while (true) {
    while (i != s->end() && (j == t->end() || *i < *j))
      s->erase(i++);
    if (i == s->end()) break;
    if (*i == *j)
      ++i;
    ++j;
  }
  return s;
}

// Change s to s intersect t.
CFGset *destructively_intersect(CFGset *s, CFGset *t)
{
  CFGset::iterator i = s->begin();
  CFGset::iterator j = t->begin();
  while (true) {
    while (i != s->end() && (j == t->end() || *i < *j))
      s->erase(i++);
    if (i == s->end()) break;
    if (*i == *j)
      ++i;
    ++j;
  }
  return s;
}

// Change s to s union t.
treeSet *destructively_union(treeSet *s, treeSet *t)
{
  s->insert( t->begin(), t->end() );
  return s;
}

treeSet *filter(bool f(const TreeNode *t), const treeSet *s)
{
  treeSet *result = new treeSet;
  remove_copy_if( s->begin(), s->end(),
		  inserter( *result, result->begin() ),
		  not1( ptr_fun( f ) ) );
  return result;
}

treeSet *destructively_filter(bool f(const TreeNode *t), treeSet *s)
{
  for (treeSet::iterator i = s->begin(); i != s->end(); ) {
    if (f(*i)) ++i;
    else s->erase(i++);
  }
  return s;
}

void printTreeSet(const treeSet *s, ostream &os, int m)
{
  if (s == NULL) {
    os << "NULL" << endl;
  } else if (s->empty()) {
    os << "empty" << endl;
  } else
    for (treeSet::const_iterator i = s->begin(); i != s->end(); ++i) {
      indent(os, m);
      os << PLCODE2(os, *i) << endl;
    }
}

void printTreeSetWithParents(const treeSet *s, ostream &os, int m)
{
  if (s == NULL) {
    os << "NULL" << endl;
  } else if (s->empty()) {
    os << "empty" << endl;
  } else
    for (treeSet::const_iterator i = s->begin(); i != s->end(); ++i) {
      indent(os, m);
      os << PLCODE2(os, *i);
      os << endl;
      indent(os, m + 1);
      os << "Par: ";
      os << PLCODE2(os, (*i)->parent());
      os << endl;
    }
}

/* Whether t appears at the top level of b. */
bool block_contains(TreeNode *t, BlockNode *b)
{
  foriter (p, b->stmts()->allChildren(), TreeNode::ChildIter)
    if (t == *p)
      return true;
  return false;
}

/* If t appears at the top level of b then return a new block that contains
   everything but t.  Otherwise return t. */
BlockNode *remove_node_from_block(TreeNode *t, BlockNode *b)
{
  bool hit = false;
  llist<TreeNode *> *l = NULL;
  foriter (p, b->stmts()->allChildren(), TreeNode::ChildIter)
    if (t == *p)
      hit = true;
    else
      push(l, *p);

  if (hit)
    return new BlockNode(dreverse(l), b->position());
  else {
    free_all(l);
    return b;
  }
}

// Is s a subset of t?
bool isSubset(const treeSet &s, const treeSet &t)
{
  return includes( t.begin(), t.end(), s.begin(), s.end() );
}

// Is s a subset of t?
bool isSubset(const treeSet *s, const treeSet *t)
{
  return isSubset(*s, *t);
}

/////////////////////////////////////////////////////////////////////////////

/* Move all VarDeclNodes in the block to the front, and make their
   initializers absent by creating separate assignment statements as needed. */
BlockNode *SortVarDeclsInBlock(BlockNode *b)
{
  TreeListNode *s = b->stmts();
  llist<TreeNode *> *decls = NULL, *newStmts = NULL;
  int c = s->arity();
  for (int i = c; i-- > 0; ) {
    TreeNode *t = s->child(i);
    if (isVarDeclNode(t)) {
      VarDeclNode *d = (VarDeclNode *) t;

      if (!d->initExpr()->absent()) {

	TreeNode *init = d->initExpr();
	d->initExpr(TreeNode::omitted);

	TreeNode *varNode = new ObjectNode (d->simpName()->clone());
	TreeNode *assn = new ExpressionStmtNode(new AssignNode(varNode, init));
	// cout << "Created initializing assignment: ";
	// cout << PCODE(assn->child(0));
	// cout << endl;
	newStmts = cons(assn, newStmts);
      }
      decls = cons(t, decls);
    } else {
      newStmts = cons(t, newStmts);
    }
  }
  newStmts = extend(decls, newStmts);
  b->stmts(new TreeListNode(newStmts));
  return b;
}

SwitchBranchNode *SortVarDeclsInSwitch(SwitchBranchNode *b)
{
  TreeListNode *s = b->stmts();
  llist<TreeNode *> *decls = NULL, *newStmts = NULL;
  int c = s->arity();
  for (int i = c; i-- > 0; ) {
    TreeNode *t = s->child(i);
    if (isVarDeclNode(t)) {
      VarDeclNode *d = (VarDeclNode *) t;

      if (!d->initExpr()->absent()) {

	TreeNode *init = d->initExpr();
	d->initExpr(TreeNode::omitted);

	TreeNode *varNode = new ObjectNode (d->simpName()->clone());
	TreeNode *assn = new ExpressionStmtNode(new AssignNode(varNode, init));
	newStmts = cons(assn, newStmts);
      }
      decls = cons(t, decls);
    } else {
      newStmts = cons(t, newStmts);
    }
  }
  newStmts = extend(decls, newStmts);
  b->stmts(new TreeListNode(newStmts));
  return b;
}

/* def appears on the list returned by use->getDefs().  Return a node
   that has the same run-time value as use, if possible.  Otherwise
   return NULL. */
TreeNode *valueTransferred(TreeNode *def, TreeNode *use)
{
  TreeNode *r = NULL;

  if (isAssignNode(def))
    r = def->opnd0();
  else if (isForEachPairNode(def))
    r = def->simpName();

  if (DEBUG_OPT) {
    cout << "Value transferred from ";
    def->pseudoprint(cout, 0);
    cout << " to ";
    use->pseudoprint(cout, 0);
    cout << ": ";
    if (r == NULL)
      cout << "?";
    else
      r->pseudoprint(cout, 0);
    cout << endl;
  }

  return r;
}

llist<TreeNode *> *add_descendants_to_list(llist<TreeNode *> *l, TreeNode *t)
{
  l = cons(t, l);
  for (int i = t->arity(); i-- > 0; )
    l = add_descendants_to_list(l, t->child(i));
  return l;
}

/* Does s contain any elt of l? */
bool set_contains_any(treeSet *s, llist<TreeNode *> *l)
{
  return find_first_of( s->begin(), s->end(),
			l->begin(), l->end() ) != s->end();
}

/* True iff no DEF of t appears in the set s. */
bool noDefsFromSet(TreeNode *t, treeSet *s)
{
  return !set_contains_any(s, t->getDefs());
}



/////////////////////////////////////////////////////////////////////////////

static treeSet * validCFG = NULL;

// If method does not have a valid CFG then do some analysis now.
// Analyses invoked: basic blocks, use/def, foreach dominators (if dominators
// is true), and deps (if computeDeps is true).
static void staticAnal(TreeNode *method, bool dominators, bool ignoreNSE,
		       bool computeDeps)
{
  garbageListClear();
  if (validCFG == NULL)
    validCFG = new treeSet;
  else if (treeSetContains(validCFG, method))
    return;
  validCFG->insert(method);

  // cout << "Setting up bblocks" << endl;
  setupBblocksMethod(method);
  // cout << "Setting up usedef" << endl;
  setupUdMethod(method, computeDeps);
  if (dominators)
    method->body()->computeForeachDominators(ignoreNSE);
  /*
  if (computeDeps)
    computeArrayAccessSummary(method);
  */
}

static void invalidateCFG()
{
  delete validCFG;
  validCFG = NULL;
}

static void redoDefUse(TreeNode *t)
{
  staticAnal(enclosingMethod(t), false);
}

/////////////////////////////////////////////////////////////////////////////

/* X should be one of PAR's children.  Make Y be PAR's child in X's place. */
void replaceChild(TreeNode *par, TreeNode *x, TreeNode *y)
{
  for (int i = par->arity(); i-- > 0; )
    if (par->child(i) == x) {
      par->child(i, y);
      return;
    }
  assert(0);
}

static void removeLazyOptNode(TreeNode *par, TreeNode *t)
{
  replaceChild(par, t, t->body());
}

static inline void renameLoops(TreeNode *t)
{
#if USE_STOPTIFU
  extern bool opt_stoptifu;
  if (opt_stoptifu) {
    /* If doing loop fusion, we need to make iteration variable names unique
       to avoid name collisions. */
    extern void loop_var_rename(TreeNode *t);
    loop_var_rename(t);
  }
#endif
}

void LazyOptimizeNode::codeGen(CodeContext &context)
{
  TreeNode *t;
  removeLazyOptNode(parent(), this);
  renameLoops(body());
  staticAnal(enclosingMethod(this), true);
  parent()->body(t = optimize(body()));
  if (DEBUG_OPT) {
    cout << "codeGen():\n\n";
    t->pseudoprint(cout, 0);
    cout << "\n\n";
  }
  t->checkIF(true);
  t->codeGen(context);
}

/* Currently unused, though it could be made to work fairly easily. */
const string LazyOptimizeNode::emitExpression(CodeContext &context)
{
  assert(0);

  TreeNode *t;
  removeLazyOptNode(parent(), this);
  renameLoops(body());
  staticAnal(enclosingMethod(this), true);
  parent()->body(t = optimize(body()));
  if (DEBUG_OPT) {
    cout << "emitExpression():\n\n";
    t->pseudoprint(cout, 0);
    cout << "\n\n";
  }
  t->checkIF(true);
  return t->emitExpression(context);
}

/////////////////////////////////////////////////////////////////////////////
// Mark parts of the tree to be optimized later.  Optimizations that require
// interprocedural information are done separately.  The LazyOptimizeNodes
// trigger opts during the codegen pass; they are also removed at that time.
/////////////////////////////////////////////////////////////////////////////

void lazyOptimize()
{
  foreach (f, llist<CompileUnitNode *>, *allFiles)
    (*f)->lazyOptimize();
}

static TreeNode *lazyOptNode(TreeNode *s)
{
  if (s == NULL || s->absent())
    return s;
  else
    return new LazyOptimizeNode(s, s->position());
}

TreeNode *TreeNode::lazyOptimize()
{
  for (int i = arity(); i-- > 0; )
    child(i, child(i)->lazyOptimize());
  return this;
}

TreeNode *MethodDeclNode::lazyOptimize()
{
  body(lazyOptNode(body()));
  return this;
}

TreeNode *ConstructorDeclNode::lazyOptimize()
{
  body(lazyOptNode(body()));
  // DOB: optimization has been disabled for constructorCall, because performing UAE there
  // currently causes live expressions to get thrown away
  // I think it's because the method def-use analysis does not include 
  // analysis of uses in constructorCall, therefore it misses those uses and axes the defs
  // there generally isn't much optimization opportunity here anyhow, so it's probably not worth fixing
  //constructorCall(lazyOptNode(constructorCall()));
  return this;
}

TreeNode *VarDeclNode::lazyOptimize()
{
  initExpr(lazyOptNode(initExpr()));
  return this;
}

/////////////////////////////////////////////////////////////////////////////

/* Dead code elimination. */
static inline TreeNode *UAE(TreeNode *t)
{
  redoDefUse(t);
  treeSet *useless = t->uselessAssignments();
  TreeNode *result = t->deleteStmts(useless);
  if (DEBUG_UAE) {
    cout << endl << "After UAE:" << endl;
    result->pseudoprint(cout, 0);
    cout << endl;
  }
  delete useless;
  return result;
}

/* Copy propagation / constant folding */
static TreeNode *CPCF(TreeNode *t, const char *s = NULL)
{
  int numchanges, totalchanges = 0;
  int iters=0;
  do {
    if (DEBUG_COPY) 
      cout << "Copy propagation / constant folding, pass #" << ++iters << endl;
    invalidateCFG();
    t->fixParent();
    redoDefUse(t);
    numchanges = doCopyPropagation(t);
    totalchanges += numchanges;
    t->foldConstants();
    if (DEBUG_COPY) 
      cout << "Made " << numchanges << " replacements." << endl;
  } while(numchanges);

  if (DEBUG_COPY) {
    if (s != NULL)
      cout << "\nCPCF pass " << s << " did " <<
	(totalchanges > 0 ? "" : "not ") << "make changes\n";
    if (totalchanges > 0) {
      cout << "\nAfter copy propagation / constant folding:\n\n";
      t->pseudoprint(cout, 0);
      cout << "\n\n";
    }
  }

  t->checkIF(true);
  return t;
}

static TreeNode * foreachOpt_and_CPCF(TreeNode *t)
{
  if (DEBUG_OPT) {
    cout << endl;
    cout << "foreachOpt:";
    PLCODE(t);
    cout << endl;
  }

  if (opt_lift || opt_sr || opt_osr) {
    t->loopAnal((TreeNode *) 0);

    /* this next line still converts foreach to foreach+ even if !opt_lift */
    t = t->liftInvariantExprs();
    checkAndFixAliasing(t, "after lifting");

    if (opt_copy)
      t = CPCF(t, "after lifting");

    if (opt_sr) {
      invalidateCFG();
      staticAnal(enclosingMethod(t), true, true);
      SRsetup();
      t = t->strengthReduce();
    }
  } else if (opt_copy)
    t = CPCF(t);

  if (DEBUG_OPT) {
    cout << endl;
    cout << "After loop opts and CPCF:";
    PLCODE(t);
    cout << endl;
  }
  checkAndFixAliasing(t, "before stoptifu");
#if USE_STOPTIFU
  if (opt_stoptifu) {
    extern TreeNode * stoptifu(TreeNode *t);
    invalidateCFG();
    staticAnal(enclosingMethod(t), true, false, true);
    t = stoptifu(t);
  }
  checkAndFixAliasing(t, "after stoptifu");
#endif
  return t;
}

TreeNode *TreeNode::applyReordering()
{
  const int childCount = arity();
  for (int sweep = 0; sweep < childCount; ++sweep)
    child(sweep, child( sweep )->applyReordering());
  
  return this;
}

TreeNode *UpdatePointBeforeStmtNode::applyReordering()
{
  if (stmt()->absent())
    stmt(deepCloneWithCrossTreeFixup(WRTloop()->stmt(), false));
  return TreeNode::applyReordering();
}

TreeNode *ReorderNode::applyReordering()
{
  TreeNode *t = stmt();
  t = ::deleteStmts(t, stmtsToDelete());
  t = deleteDeclsFromBlocks(t, declsToDelete());
  {
    map_tree_to_tree &m = *(remap());
    for (map_tree_to_tree::iterator i = m.begin(); i != m.end(); i++)
      t = replaceOneNode(t, i->first, i->second);
  }
  // t = t->collapseTrivialBlocks();
  t->fixParent();
  delete this;
  return t->applyReordering();
}

/*
   Given a piece of code to be optimized, optimize() returns an
   optimized version of that code.  The input may be destructively
   modified in the process.  (Which optimizations to perform is controlled
   by command line flags.)
*/
static TreeNode *optimize(TreeNode *t)
{
  debug_lift = DEBUG_LIFT;
  debug_sr = DEBUG_SR;

  if (DEBUG_OPT) {
    cout << "Optimizing:\n\n";
    t->pseudoprint(cout, 0);
    cout << "\n\n";
  }

  t = foreachOpt_and_CPCF(t);

  /* currently unnecessary
  if (opt_copy)
    t = CPCF(t, "before UAE");
  */

  if (opt_uae) {
    invalidateCFG();
    t->fixParent();
    t = UAE(t);
  }

  if (opt_unreachable) {
    t->fixParent();
    setupBblocksMethod(enclosingMethod(t));    
    removeUnreachableStmts(t);
  }

  if (opt_deadvar) {
    t->fixParent();
    t = doDeadVariableRemoval(t);
  }

  invalidateCFG();
  t->fixParent();
  t = t->applyReordering();
  t = t->collapseTrivialBlocks();

  if (DEBUG_OPT) {
    cout << "Result of optimization:\n\n";
    t->pseudoprint(cout, 0);
    cout << "\n\n";
  }

  checkAndFixAliasing(t, "after optimization");
  return t;
}

