#include "backend.h"
#include "ti-gc.h"
#include <assert.h>

/* Statistical collection and reporting for AMUDP/AMMPI backends */
#if defined(COMM_AMUDP) || defined(COMM_AMMPI)
#define CONCAT(x,y) x ## y
#ifdef COMM_AMUDP
 #define P(fnname) CONCAT(AMUDP_,fnname)
 #define p(fnname) CONCAT(amudp_,fnname)
#else
 #define P(fnname) CONCAT(AMMPI_,fnname)
 #define p(fnname) CONCAT(ammpi_,fnname)
#endif

static int statscalls = 0;
static p(stats_t) globalStats;

void ___amstats_request_handler(void *token, void *buf, int nbytes, int procnum, int reserved) {
  assert(nbytes == sizeof(p(stats_t)));
  AM_Safe(P(AggregateStatistics)(&globalStats, (p(stats_t) *)buf));
  statscalls++;
  }
static void printGlobalStats() {
  p(stats_t) stats;
  statscalls = 0; 
  globalStats = P(initial_stats);

  /* make sure we're done sending msgs for now */
  barrier();

  if (MYBOXPROC == 0) {
    if (usingAMSPMD) { P(SPMDBarrier)(); /* try to minimize impact of monitoring */
      #ifdef COMM_AMMPI
        { /* problem - AMMPI uses messages to implement above barrier -
           * make sure P0 gets all the barrier replies, or above counts won't match up
           */
          ___tic_time_t tm;
          ___tic_set_wait_timer(&tm, 100); /* spin-poll for 100 milliseconds */
          while (___tic_millisuntil(&tm) >= 0) {
            tic_poll(); 
            tic_sched_yield();
          }
        }
      #endif
    }

    AM_Safe(P(GetEndpointStatistics)(CUR_ENDPOINT, &stats)); /* get statistics */

    /* don't let stats msgs interfere with display */
    if (usingAMSPMD) P(SPMDBarrier)(); 

    assert(sizeof(p(stats_t)) < AM_MaxMedium());
    TI_AM_RequestI2(CUR_ENDPOINT, 0, __AMSTATS_H, &stats, sizeof(p(stats_t)), MYBOX, 0); /* send to zero */

    if (MYPROC == 0) {
      while (statscalls < BOXES) {
        tic_poll(); 
        tic_sched_yield();
        }

      fprintf(stderr, "--------------------------------------------------\n"
                      "Global AM2 usage statistics:\n");
      AM_Safe(P(DumpStatistics)(stderr, &globalStats, 1));
      fprintf(stderr, "--------------------------------------------------\n");
      fflush(stderr);

      sleep(1); /* HACK: give a little time for this output to reach user's console before shutdown */
      }
    }

  barrier(); /* just to keep things clean */

  }
#else
static void printGlobalStats() {
  }
#endif

static int TI_GCSTATS = 0;
static int TI_PAGESTATS = 0;
static int TI_AMSTATS = 0;

int dynamicCallCount_staticNative = 0;
int dynamicCallCount_staticSpecial = 0;
int dynamicCallCount_staticTitanium = 0;
int dynamicCallCount_dynamic = 0;

void stats_init(void) {
  /* read environment info at startup, because it may not be available later (AMUDP backends) */
  TI_GCSTATS = !!getenvMaster("TI_GCSTATS");
  TI_PAGESTATS = !!getenvMaster("TI_PAGESTATS");
  TI_AMSTATS = !!getenvMaster("TI_AMSTATS");
}

void print_page_stats(char *header);

/* printExitStats(int sglobal):
 * should be called whenever exiting a Titanium program via any means
 * handles the output of any requested runtime statistics
 * sglobal indicates whether all processors are running this and sglobal operations are permitted 
 */
void printExitStats(int sglobal) {

  if (MYBOXPROC == 0 || !sglobal) {
    /* exit stats which should run once per box */
    char node_identifier[40];

    if (MYBOXPROCS > 1) { 
      int firstproc = COMM_GetProxyProcNumberForBoxNumber(MYBOX);
      sprintf(node_identifier,"Box %i (P%i..P%i)", 
              MYBOX, firstproc, firstproc + MYBOXPROCS-1);         
    }
    else sprintf(node_identifier,"P%i", MYPROC);

    #ifndef USE_GC_NONE
    if (TI_GCSTATS) {
      printf("%s: Heap size: %d bytes.\n", node_identifier, ti_gc_usedheapsize());
      #if defined(USE_DISTRIBUTED_GC) 
        printf("%s: Escaped global pointers (upper bound): %i\n", node_identifier, dgc_numhashed());
      #endif
    }
    #endif

    if (TI_PAGESTATS) print_page_stats(node_identifier);

    if (dynamicCallCount_staticNative   || 
        dynamicCallCount_staticSpecial  ||
        dynamicCallCount_staticTitanium ||
        dynamicCallCount_dynamic) {
      printf("%s: Method call stats: %i native, %i native-special, %i non-virtual, %i virtual\n", 
        node_identifier, 
        dynamicCallCount_staticNative,
        dynamicCallCount_staticSpecial,
        dynamicCallCount_staticTitanium,
        dynamicCallCount_dynamic);
    }
  }

  #if defined(COMM_AMUDP) || defined(COMM_AMMPI)
    if (usingAMSPMD && TI_AMSTATS && COMM_GetBParallelDegree() > 1) {
      if (!sglobal) {  /* we can't print stats here because it's an sglobal operation */
        fprintf(stderr, "System.exit() called - unable to print AM stats");
        fflush(stderr);
      } else { /* all boxes present - print AMUDP stats */
        printGlobalStats();
      }
    }
  #endif
}

