#include <assert.h>
#include <tic.h>
#include "ti-gc.h"

/* TODO: the AM message arguments in this file will need to be fixed up 
 *  if we ever migrate to a 64-bit AM implementation
 */

/**************************
 * INITIALIZATION/STATICS *
 *************************/

#ifdef COMM_AM2
static handler_t packHandlerIdx, unpackAllHandlerIdx, unpackOnlyHandlerIdx;
static handler_t packReplyHandlerIdx, unpackReplyHandlerIdx;
static handler_t deleteHandlerIdx, nullReplyHandlerIdx, allocHandlerIdx,
                 allocReplyHandlerIdx;

typedef void *(*pack_method_t)(void *, int *, void *);
typedef void *(*unpack_method_t)(void *, void *);

static void packHandler(void *token, void *copy_desc, int copy_desc_size,
			int _packMethod, int remoteBuffer, int pack_info_ptr, 
			int atomicelements);
static void packReplyHandler(void *token, void *array_data, 
			     int array_data_size, int localBuffer,
			     int remoteBuf, int dataSz,
                             int pack_info_ptr);
static void unpackAllHandler(void *token, void *packedArrayData, int nBytes, 
		             int _unpackMethod, int copyDescSize, int unpack_spin_ptr, 
		             int dummy3);
static void unpackOnlyHandler(void *token, void *copyDesc, int copyDescSize,
			      int bufAddr, int _unpackMethod, int unpack_spin_ptr,
			      int dummy3);
static void unpackReplyHandler(void *token, int unpack_spin_ptr, 
                               int dummy1, int dummy2, int dummy3);
static void deleteHandler(void *token, int addr, int dummy1, int dummy2,
			  int dummy3);
static void allocHandler(void *token, int size, int destptr, int dummy2,
			  int dummy3);
static void allocReplyHandler(void *token, int remAddr, int destptr, 
			      int dummy2, int dummy3);
static void nullReplyHandler(void *token);

/* sparse scatter-gather */
static handler_t simpleScatterHandlerIdx, generalScatterHandlerIdx, doneReplyHandlerIdx;

static handler_t simpleGatherHandlerIdx, simpleGatherReplyHandlerIdx,
                 generalGatherHandlerIdx;

static void simpleScatterHandler(void *token, void *addr_data_list, int addr_data_list_size,
			      int num_elem, int elem_sz, 
                              int _done_ctr, int dummy1);
static void generalScatterHandler(void *token, int addr_data_list,
                                  int num_elem, int elem_sz, 
                                  int _done_ctr);
static void doneReplyHandler(void *token, int _done_ctr);
static void simpleGatherHandler(void *token, void *_addr_list, int addr_list_size,
			      int num_elem, int elem_sz, 
                              int _tgt_data_list, int _done_ctr,
                              int atomic_elements,
                              int dummy1, int dummy2, int dummy3);
static void simpleGatherReplyHandler(void *token, void *data_list, int data_list_size,
			      int num_elem, int elem_sz, 
                              int _tgt_data_list, int _done_ctr);
static void generalGatherHandler(void *token, int addr_data_list,
                                  int num_elem, int elem_sz, 
                                  int _done_ctr);

void bulkSetupHandlers() {
  if (COMM_GetBParallelDegree() > 1) {
    /* choose any available handler index for these handlers */
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &packHandlerIdx, packHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &unpackAllHandlerIdx, unpackAllHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &unpackOnlyHandlerIdx, unpackOnlyHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &packReplyHandlerIdx, packReplyHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &unpackReplyHandlerIdx, unpackReplyHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &deleteHandlerIdx, deleteHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &allocHandlerIdx, allocHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &allocReplyHandlerIdx, 
			     allocReplyHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &nullReplyHandlerIdx, nullReplyHandler));

    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &simpleScatterHandlerIdx, simpleScatterHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &generalScatterHandlerIdx, generalScatterHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &doneReplyHandlerIdx, doneReplyHandler));

    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &simpleGatherHandlerIdx, simpleGatherHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &simpleGatherReplyHandlerIdx, simpleGatherReplyHandler));
    AM_Safe(AM_SetHandlerAny(__sc_endpoint, &generalGatherHandlerIdx, generalGatherHandler));
  }
}
/* ------------------------------------------------------------------------------------ */
/***************************
 * Miscellaneous handlers. *
 ***************************/

static void nullReplyHandler(void *token) {
  return;
}

static void deleteHandler(void *token, int addr, int dummy1, int dummy2,
			  int dummy3) {
  ti_free((void *)addr);
  TI_AM_Reply0(token, nullReplyHandlerIdx);
}

/* Size is in bytes. */
static void allocHandler(void *token, int size, int destptr, int dummy2,
			  int dummy3) {
#ifdef USE_GC_NONE
  void *buf = (void *)ti_malloc(size);
#else
  /* need to ensure it won't be collected since no local ptrs remain */
  void *buf = (void *)GC_malloc_uncollectable(size); 
#endif
  TI_AM_Reply4(token, allocReplyHandlerIdx, (int)buf, destptr, 0, 0);
}

static void allocReplyHandler(void *token, int buf, int destptr, int dummy2,
			      int dummy3) {
  void ** premoteAllocBuf = (void **)destptr;
  (*premoteAllocBuf) = (void *)buf;
}

/* ------------------------------------------------------------------------------------ *
 *  Rectangular scatter-gather
 * ------------------------------------------------------------------------------------ */
/************************************
 * Get an array from a remote node. *
 ***********************************/

/* Call _packMethod to pack the array with copy_desc as its "copy descriptor",
   returning the packed data to remote address remoteBuffer. */
static void packHandler(void *token, void *copy_desc, int copy_desc_size,
			int _packMethod, int remoteBuffer, int pack_info_ptr, 
			int atomicelements) {
  int array_data_len;
  void *array_data;
  pack_method_t packMethod = (pack_method_t)_packMethod;
  /* A NULL buffer pointer tells the packer to allocate its own buffer. 
   * (void*)-1 tells it to use uncollectable memory so it doesn't disappear on us
   */
  array_data = (*packMethod)(copy_desc, &array_data_len, (void*)-1);

  #if defined(USE_DISTRIBUTED_GC) && !defined(USE_GC_NONE)
    /* inform distributed GC that some pointers may be escaping */
    if (!atomicelements)
      GC_PTR_ESC_ALL_GPS(array_data, array_data_len);
  #endif

  if (array_data_len <= ___AM_MAX_MEDIUM) {
    /* Common, fast case. Send back the packed data with the reply. */
    TI_AM_ReplyI4(token, packReplyHandlerIdx, array_data, 
		       array_data_len, remoteBuffer, 0, 0, pack_info_ptr);
    ti_free(array_data);
  }
  else {
    /* Slower case. Why doesn't the darned AM_ReplyXfer() work? */
    TI_AM_ReplyI4(token, packReplyHandlerIdx, array_data,
		       ___AM_MAX_MEDIUM, remoteBuffer, (int)array_data, 
		       array_data_len, pack_info_ptr);
  }
}

/* The reply handler needs to pass this information back to the caller. */
typedef struct {
  volatile int pack_spin; /* 0 while waiting, 1 if completed, 2 if using fragmentation */
  void * volatile remoteBuffer;  
  volatile int dataSize;  
} pack_return_info_t;

/* Sets the spin flag to 1 if the array fit in the message reply and we're
   done; otherwise it sets the spinlock variable to be the remote address 
   of the packed data, which must be sent over with a libtic bulk 
   transfer. */
static void packReplyHandler(void *token, void *array_data, 
			     int array_data_size, int localBuffer,
			     int remoteBuf, int dataSz,
                             int pack_info_ptr) {
  int moreWaiting = (remoteBuf != 0 && dataSz != 0);
  pack_return_info_t* pinfo = (pack_return_info_t*)pack_info_ptr;
  memcpy((void *)localBuffer, array_data, array_data_size);
  if (!moreWaiting) { /* got it all in a single chunk */
    pinfo->pack_spin = 1;
  }
  else { /* need more chunks */
    assert(array_data_size == ___AM_MAX_MEDIUM); /* first chunk always max medium in size */
    pinfo->remoteBuffer = (jbyte *)remoteBuf;
    pinfo->dataSize = dataSz;
    pinfo->pack_spin = 2;
  }
}

void get_array(void *pack_method, void *copy_desc, int copy_desc_size, 
	       int tgt_box, void *buffer, int atomicelements) {
  pack_return_info_t info;
  info.pack_spin = 0;
  /* A copy_desc is an object that contains the array descriptor. */
  assert(copy_desc && copy_desc_size > 0 && copy_desc_size <= ___AM_MAX_MEDIUM);
  TI_AM_RequestI4(__sc_endpoint, tgt_box, packHandlerIdx, 
		       copy_desc, copy_desc_size, (int)pack_method, 
		       (int)buffer, (int)&info, (int)atomicelements);
  while (!info.pack_spin) { tic_poll(); }
  /* If the data was too big to be sent in the reply, use a Tic call
     to do the job. */
  if (info.pack_spin != 1) {
    /* pull the data to use using a bulk copy (first fragment has already been copied) */
    jGPointer remoteBufPtr = toglobalb(tgt_box, ((jbyte *)info.remoteBuffer) + ___AM_MAX_MEDIUM);
    bulk_read((void *)((int)buffer+___AM_MAX_MEDIUM), remoteBufPtr, info.dataSize - ___AM_MAX_MEDIUM, 
        (atomicelements?tic_no_ptrs:tic_gp_only));
    /* now tell remote to delete temp space */
    TI_AM_Request4(__sc_endpoint, tgt_box, deleteHandlerIdx, 
			(int)info.remoteBuffer, 0, 0, 0);
  }
}

/************************************
 * Write an array to a remote node. *
 ***********************************/

/* The fastest way of doing things....take the data given and unpack it. */
static void unpackAllHandler(void *token, void *packedArrayData, int nBytes, 
		   int _unpackMethod, int copyDescSize, int unpack_spin_ptr, 
		   int dummy3) {
 void *temp = NULL;
 if (((int)packedArrayData) % 8 != 0) {
   /* it seems that some crappy AM implementations (ahem.. the NOW)
    * fail to double-word align the medium-sized message buffer
    * fix that problem if it occurs
    */
   temp = ti_malloc(nBytes+8);
   memcpy(temp, packedArrayData, nBytes);
   packedArrayData = temp;
 } 
 {
   unpack_method_t unpackMethod = (unpack_method_t)_unpackMethod;
   void *copyDesc = packedArrayData;
   void *arrayData = (void *)((int)packedArrayData + copyDescSize);
   assert(((int)copyDesc) % 8 == 0 && ((int)arrayData) % 8 == 0);
   (*unpackMethod)(copyDesc, arrayData);
   TI_AM_Reply4(token, unpackReplyHandlerIdx, unpack_spin_ptr, 0, 0, 0);
 }
 if (temp) ti_free(temp);
}

static void unpackReplyHandler(void *token, int unpack_spin_ptr, 
                               int dummy1, int dummy2, int dummy3) {
  int *punpack_spin = (int *)unpack_spin_ptr;
  *(punpack_spin) = 1;
}

/* The data was provided by previous calls; just unpack it with the given
   descriptor. */
static void unpackOnlyHandler(void *token, void *copyDesc, int copyDescSize,
			      int bufAddr, int _unpackMethod, int unpack_spin_ptr,
			      int dummy3) {
  unpack_method_t unpackMethod = (unpack_method_t)_unpackMethod;
  (*unpackMethod)(copyDesc, (void *)bufAddr);
  ti_free((void *)bufAddr);
  TI_AM_Reply4(token, unpackReplyHandlerIdx, unpack_spin_ptr, 0, 0, 0);
}
  

/* Send a contiguous array of data to a node, and have it get unpacked into
   a Titanium array. */
void put_array(void *unpack_method, void *copy_desc, int copy_desc_size,
	       void *array_data, int array_data_size, int tgt_box) {
  void *data;
  /* ensure double-word alignment for array data */
  int copy_desc_size_padded = ((copy_desc_size-1)/8 + 1) * 8; 
  int data_size = copy_desc_size_padded + array_data_size;
  volatile int unpack_spin = 0;

  /* Fast(er), hopefully common case. */
  if (data_size <= ___AM_MAX_MEDIUM) {
    /* The remote end needs the descriptor and the array data. Since the
       data is small, copying the array data will be faster than
       sending two messages. */
    data = (void *) ti_malloc(data_size);
    assert(data && copy_desc_size_padded % 8 == 0);
    memcpy(data, copy_desc, copy_desc_size);
    /* All math is done in bytes. sizeof(void*) should be irrelevant. */
    memcpy((void *)((int)data + copy_desc_size_padded), array_data, array_data_size);
    TI_AM_RequestI4(__sc_endpoint, tgt_box, unpackAllHandlerIdx, 
			 data, data_size, (int)unpack_method, 
			 copy_desc_size_padded, (int)&unpack_spin, 0);
    while (!unpack_spin) { tic_poll(); }
    ti_free(data);
  }
  else { /* Slow case. */
    jGPointer remotePtr;
    /* Allocate a buffer to hold the array data on the remote side. */
    volatile void *remoteAllocBuf = NULL;
    TI_AM_Request4(__sc_endpoint, tgt_box, allocHandlerIdx, 
			array_data_size, (int)&remoteAllocBuf, 0, 0);
    while (!remoteAllocBuf) { tic_poll(); }
    /* Transfer the data to the buffer with libtic. */
    remotePtr = toglobalb(tgt_box, (void *)remoteAllocBuf);
    bulk_write(remotePtr, array_data, array_data_size);
    /* Tell the remote side to unpack the data. */
    assert(copy_desc_size <= ___AM_MAX_MEDIUM);
    TI_AM_RequestI4(__sc_endpoint, tgt_box, unpackOnlyHandlerIdx,
			 copy_desc, copy_desc_size, (int)remoteAllocBuf, 
			 (int)unpack_method, (int)&unpack_spin, 0);
    while (!unpack_spin) { tic_poll(); }
  }
}
/* ------------------------------------------------------------------------------------ *
 *  Sparse scatter-gather
 * ------------------------------------------------------------------------------------ */
/* this macro requires dest and src have alignment of at least length */
#define FAST_MEMCPY(dest, src, length) do {           \
  switch(length) {                                    \
    case sizeof(jubyte):                              \
      *((jubyte *)(dest)) = *((jubyte *)(src));       \
      break;                                          \
    case sizeof(jchar):                               \
      *((jchar *)(dest)) = *((jchar *)(src));         \
      break;                                          \
    case sizeof(juint):                               \
      *((juint *)(dest)) = *((juint *)(src));         \
      break;                                          \
    case sizeof(julong):                              \
      *((julong *)(dest)) = *((julong *)(src));       \
      break;                                          \
    default:                                          \
      memcpy(dest, src, length);                      \
  } } while(0)

void sparse_scatter(void **remote_addr_list, void *src_data_list, 
                    int remote_box, int num_elem, int elem_sz, int atomic_elements) {
  volatile int done_ctr = 0;
  int datasz;
  int offset;

  assert(src_data_list && remote_addr_list && 
         remote_box >= 0 && remote_box < BOXES &&
         num_elem > 0 && elem_sz > 0);

  offset = num_elem * sizeof(void *);
  datasz = offset + num_elem * elem_sz;
  if (datasz <= ___AM_MAX_MEDIUM) { 
    /* fast case - everything fits in a medium msg */
    char *data = ti_malloc(datasz);
    memcpy(data, remote_addr_list, offset);
    memcpy(data+offset, src_data_list, num_elem * elem_sz);
    #if defined(USE_DISTRIBUTED_GC) && !defined(USE_GC_NONE)
      /* inform distributed GC that some pointers may be escaping */
      if (!atomic_elements) 
        GC_PTR_ESC_ALL_GPS(data+offset, num_elem * elem_sz);
    #endif
    TI_AM_RequestI4(__sc_endpoint, remote_box, simpleScatterHandlerIdx,
			 data, datasz, 
                         num_elem, elem_sz,
                         (int)&done_ctr, 0);
    while (!done_ctr) { tic_poll(); }
    ti_free(data);
  } else {  /* slower case - need multiple messages */
    jGPointer remotePtr;
    volatile void *remoteAllocBuf = NULL;

    /* Allocate a buffer to hold the array data on the remote side. */
    TI_AM_Request4(__sc_endpoint, remote_box, allocHandlerIdx, 
			datasz, (int)&remoteAllocBuf, 0, 0);
    while (!remoteAllocBuf) { tic_poll(); }

    /* Transfer the addresses and data to the buffer with libtic (this could be optimized somewhat) */
    remotePtr = toglobalb(remote_box, (void *)remoteAllocBuf);
    bulk_write(remotePtr, remote_addr_list, offset);
    #if defined(USE_DISTRIBUTED_GC) && !defined(USE_GC_NONE) & 0
      /* inform distributed GC that some pointers may be escaping */
      if (!atomic_elements) 
        GC_PTR_ESC_ALL_GPS(src_data_list, num_elem * elem_sz);
    #endif
    remotePtr = toglobalb(remote_box, ((char *)remoteAllocBuf)+offset);
    bulk_write(remotePtr, src_data_list, num_elem * elem_sz);

    /* Tell the remote side to scatter the data. */
    TI_AM_Request4(__sc_endpoint, remote_box, generalScatterHandlerIdx,
			 (int)remoteAllocBuf, num_elem, elem_sz, (int)&done_ctr);
    while (!done_ctr) { tic_poll(); }
  }
}
static void simpleScatterHandler(void *token, void *addr_data_list, int addr_data_list_size,
			      int num_elem, int elem_sz, 
                              int _done_ctr, int dummy1) {
  int i;
  void **addr_list = (void**)addr_data_list;
  int offset = num_elem * sizeof(void *);
  char *data = ((char*)addr_data_list) + offset;
  assert(addr_data_list_size == offset + num_elem * elem_sz);
  for (i = 0; i < num_elem; i++) {
    FAST_MEMCPY(addr_list[i], data + i*elem_sz, elem_sz); 
  }
  TI_AM_Reply1(token, doneReplyHandlerIdx, _done_ctr);
} 
static void doneReplyHandler(void *token, int _done_ctr) {
  int *done_ctr = (int *)_done_ctr;
  *done_ctr = 1;
}
static void generalScatterHandler(void *token, int addr_data_list,
                                  int num_elem, int elem_sz, 
                                  int _done_ctr) {
  int i;
  void **addr_list = (void**)addr_data_list;
  int offset = num_elem * sizeof(void *);
  char *data = ((char*)addr_data_list) + offset;
  for (i = 0; i < num_elem; i++) {
    FAST_MEMCPY(addr_list[i], data + i*elem_sz, elem_sz); 
  }
  TI_AM_Reply1(token, doneReplyHandlerIdx, _done_ctr);
  ti_free((void*)addr_data_list);
} 
/* ------------------------------------------------------------------------------------ */
void sparse_gather(void *tgt_data_list, void **remote_addr_list, 
                   int remote_box, int num_elem, int elem_sz, int atomic_elements) {
  volatile int done_ctr = 0;

  assert(tgt_data_list && remote_addr_list && 
         remote_box >= 0 && remote_box < BOXES &&
         num_elem > 0 && elem_sz > 0);

  if (num_elem * sizeof(void *) <= ___AM_MAX_MEDIUM &&
      num_elem * elem_sz <= ___AM_MAX_MEDIUM) { /* fast case - everything fits in a medium msg */
    TI_AM_RequestI8(__sc_endpoint, remote_box, simpleGatherHandlerIdx,
			 remote_addr_list, num_elem * sizeof(void *), 
                         num_elem, elem_sz,
                         (int)tgt_data_list, (int)&done_ctr,
                         (int)atomic_elements, 
                         0,0,0);
    while (!done_ctr) { tic_poll(); }
  } else {  /* slower case - need multiple messages */
    int offset = num_elem * sizeof(void *);
    int datasz = offset + num_elem * elem_sz;
    jGPointer remotePtr;
    volatile void *remoteAllocBuf = NULL;

    /* Allocate a buffer to hold the array data on the remote side. */
    TI_AM_Request4(__sc_endpoint, remote_box, allocHandlerIdx, 
			datasz, (int)&remoteAllocBuf, 0, 0);
    while (!remoteAllocBuf) { tic_poll(); }

    /* Transfer the addresses to the buffer with libtic (this could be optimized somewhat) */
    remotePtr = toglobalb(remote_box, (void *)remoteAllocBuf);
    bulk_write(remotePtr, remote_addr_list, offset);

    /* Tell the remote side to gather the data. */
    TI_AM_Request4(__sc_endpoint, remote_box, generalGatherHandlerIdx,
			 (int)remoteAllocBuf, num_elem, elem_sz, (int)&done_ctr);
    while (!done_ctr) { tic_poll(); }

    /* Transfer the data from the buffer with libtic (this could be optimized somewhat) */
    remotePtr = toglobalb(remote_box, ((char *)remoteAllocBuf)+offset);
    bulk_read(tgt_data_list, remotePtr, num_elem * elem_sz, 
        (atomic_elements?tic_no_ptrs:tic_gp_only)); /* (handles gp escape too) */

    TI_AM_Request4(__sc_endpoint, remote_box, deleteHandlerIdx, 
			(int)remoteAllocBuf, 0, 0, 0);
  }
}
static void simpleGatherHandler(void *token, void *_addr_list, int addr_list_size,
			      int num_elem, int elem_sz, 
                              int _tgt_data_list, int _done_ctr,
                              int atomic_elements,
                              int dummy1, int dummy2, int dummy3) {
  int i;
  void **addr_list = (void**)_addr_list;
  int datasz = num_elem * elem_sz;
  char *data = ti_malloc(datasz);
  assert(addr_list_size == sizeof(void *)*num_elem);
  for (i = 0; i < num_elem; i++) {
    FAST_MEMCPY(data + i*elem_sz, addr_list[i], elem_sz); 
  }
  #if defined(USE_DISTRIBUTED_GC) && !defined(USE_GC_NONE) && 0
    /* inform distributed GC that some pointers may be escaping */
    if (!atomic_elements)
      dgc_ptr_esc_all_gps(data, datasz, 1);
  #endif
  TI_AM_ReplyI4(token, simpleGatherReplyHandlerIdx,
		       data, datasz, 
                       num_elem, elem_sz,
                       _tgt_data_list, _done_ctr);
  ti_free(data);
} 
static void simpleGatherReplyHandler(void *token, void *data_list, int data_list_size,
			      int num_elem, int elem_sz, 
                              int _tgt_data_list, int _done_ctr) {
  void *tgt_data_list = (void *)_tgt_data_list;
  int *done_ctr = (int *)_done_ctr;
  assert(data_list_size == num_elem * elem_sz);
  memcpy(tgt_data_list, data_list, data_list_size);
  *done_ctr = 1;
}
static void generalGatherHandler(void *token, int addr_data_list,
                                  int num_elem, int elem_sz, 
                                  int _done_ctr) {
  int i;
  void **addr_list = (void**)addr_data_list;
  int offset = num_elem * sizeof(void *);
  char *data = ((char*)addr_data_list) + offset;
  for (i = 0; i < num_elem; i++) {
    FAST_MEMCPY(data + i*elem_sz, addr_list[i], elem_sz); 
  }
  TI_AM_Reply1(token, doneReplyHandlerIdx, _done_ctr);
} 
/* ------------------------------------------------------------------------------------ */

#endif /* COMM_AM2 */
