/* tic.h - main header file for tic backends */
/* see copyright.txt for usage terms */

#ifndef TIC_H
#define TIC_H

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>

#if 0
#ifndef __GNUC__
#include <ansi-c.h>
#endif
#endif

/* .............................
 *
 * Control `extern' and `inline'
 *
 * .............................
 */


/* Usage: 
 * The proper (portable) way to declare inline functions in tic headers is:
 *
 *   INLINE_MODIFIER int foo(int bar) 
 *   INLINE ({
 *     function body...
 *     })
 */

/* start with a clean slate */
#ifdef EXTERN
#undef EXTERN
#endif
#ifdef INLINE
#undef INLINE
#endif

/* Defining TIC_DEBUG has the effect of disabling all inlining of functions
 * which helps when debugging. */
#if defined(TIC_DEBUG)
#	define	INLINE(code) ;
#else
#	define	INLINE(code) {code}
#endif

/* Defining TIC_INTERNAL_DEFINE has the effect of producing a
 * real (non-extern) version of all functions and variables. 
 *
 * Exactly one .c file should define this in the distribution to 
 * support the TIC_DEBUG option and non-GNU compilers
 */
#ifdef TIC_INTERNAL_DEFINE
#      define EXTERN
#      undef INLINE
#      define INLINE(code) {code}
#      define FORCE_NO_RUNTIME_INLINE
#      define FORCE_STATIC 0
#else
#      define EXTERN extern
#      define FORCE_STATIC 1
#endif

/* get definition for INLINE_MODIFIER, INLINE_MODIFIER_FOR_SOLARIS_AM */
#include "runtime-options-inl.h"


#include "backend-defines.h"
#include <primitives.h>
#include "gp-type.h"


/* ............................
 *
 * Control debugging statements
 *
 * ............................
 */

#ifdef TIC_DEBUG
#define SC_DEBUG(a) a
#else
#define SC_DEBUG(a)
#endif

/* ..........
 *
 * Boxes and Processors
 *
 * ..........
 */

#define MAX_BOX_PROCS 1024
#define MAX_PROCS 1024

#ifdef MEMORY_SHARED
#define ___MYPROC       myProcessNumber()
#define ___MYBOXPROC    myBoxProcessNumber()
extern int ___MYBOXPROCS;
#else
extern Process ___MYPROC;
#define ___MYBOXPROC    0
#define ___MYBOXPROCS  	1
#endif
extern int ___PROCS;
extern int ___BOXES;
extern Box ___MYBOX;

/* extern __tic_thread_key_t thread_key; */

#define MYPROC     ___MYPROC
#define PROCS      ___PROCS
#define MYBOX      ___MYBOX
#define MYBOXPROC  ___MYBOXPROC
#define MYBOXPROCS ___MYBOXPROCS
#define BOXES      ___BOXES

extern int ___AM_MAX_SHORT; /* size of short packet supported by AM */
extern int ___AM_MAX_MEDIUM; /* size of short packet supported by AM */
extern int ___AM_MAX_LONG; /* size of short packet supported by AM */

extern void *tic_main(int argc, char **argv);

/* ..........
 *
 * GC stuff - MUST come before pthread stuff
 *
 * ..........
 */


#ifndef USE_GC_NONE
#include "gc/gc.h"
#endif

/* ..........
 *
 * Thread operation wrappers
 *
 * ..........
 */

#define EXTRA_SYNC_CHECKING 0 /* set to non-zero to enable extra checking for error values */


#if defined (THREAD)  /* Solaris Threads */
#include <thread.h>

#define __tic_thread_t thread_t
#define __tic_thread_self() thread_self()
#define __tic_create_thread(func, arg, threadptr, result)  \
  do {   \
    result = thr_create(NULL, NULL, func, arg, THR_BOUND, threadptr); \
  } while(0) 
#define __tic_thread_join(thread, retval)  thr_join(0, NULL, retval)
#define __tic_thread_set_concurrency(conc) thr_setconcurrency(conc)

#define __tic_thread_key_t thread_key_t
#define __tic_thread_key_init(key)     thr_keycreate(key, NULL)
#define __tic_thread_get_key_value(key, ptr)  thr_getspecific(key, (void **)ptr)
#define __tic_thread_set_key_value(key, ptr)  thr_setspecific(key, (void *)(ptr))

#elif defined(PTHREAD)

#include <pthread.h>

#define __tic_thread_t pthread_t
#define __tic_thread_self() pthread_self()
#define __tic_create_thread(func, arg, threadptr, result)  \
  do {    \
    pthread_attr_t attr;    \
    pthread_attr_init(&attr);   \
    pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM); \
    result = pthread_create(threadptr, &attr, func, arg);    \
  } while (0)
#define __tic_thread_join(thread, retval)  pthread_join(thread, retval)
#ifdef HAVE_PTHREAD_SETCONCURRENCY
/* this isn't supported in all pthread implementations - 
   however, it's just a hint to the pthread scheduler anyhow */
  #define __tic_thread_set_concurrency(conc) pthread_setconcurrency(conc)
#else
  #define __tic_thread_set_concurrency(conc) 
#endif

#define __tic_thread_key_t pthread_key_t 
#define __tic_thread_key_init(key)     pthread_key_create(key, NULL)
#define __tic_thread_get_key_value(key, ptr)  (*ptr) = pthread_getspecific(key)
#define __tic_thread_set_key_value(key, ptr)  pthread_setspecific(key, ptr)

#else /* NO Threads */

#define __tic_thread_t  int
#define __tic_thread_self() 0
#define __tic_create_thread(func, arg, threadptr, result) (result=0)
#define __tic_thread_join(thread, retval)  (retval?(*(void**)retval)=NULL,0:0)
#define __tic_thread_set_concurrency(conc) 0

#define __tic_thread_key_t void **
#define __tic_thread_key_init(key)  ((*key) = (void **)ti_malloc(sizeof(void *)), (int)!(*key))
#define __tic_thread_get_key_value(key, ptr)  ((*(void **)ptr) = (*key), 0)
#define __tic_thread_set_key_value(key, ptr)  ((*key) = (void *)ptr, 0)


#endif /* Threads */


/* Locks and Mutexes */

#if defined(IRIX_THREADS) 
#define CAUTIOUS_MONITORS
#endif

#ifndef MEMORY_SHARED

#define __tic_lock_t int
#define __tic_cond_t int

#define __tic_lock_init(lock) 0     
#define __tic_lock_initializer  0
#define __tic_lock_decl_initializer 0
#define __tic_cond_init(cond, lock)
#define __tic_cond_initializer  0
#define __tic_cond_decl_initializer 0

#define __tic_lock(lock) 0 
#define __tic_try_lock(lock) 1          
#define __tic_unlock(lock) 0         
#define __tic_mon_wait(mon)           abort()
#define __tic_mon_timedwait(mon, tm)  abort()
#define __tic_mon_signal(mon)         abort()
#define __tic_mon_broadcast(mon)      abort()

#define __tic_lock_destroy(lock) 0
#define __tic_cond_destroy(cond) 0

#define ___tic_time_t	struct timespec

#else

/* error checking macros for thread functions */
#if EXTRA_SYNC_CHECKING
  INLINE_MODIFIER int __tic_errcheck_helper(int actualval, int okval, int altok, char* where) 
  INLINE({                                  
    if ((!!actualval) == okval || actualval == altok) return actualval;
    else {
      char altstr[20];
      if (altok != okval) sprintf(altstr, "(or %i)", altok);
      else altstr[0] = '\0';
      fprintf(stderr, "Tic: Error while calling: %s\n", where);  
      fprintf(stderr, "Tic: expected %s%s, got %i\n", 
        (okval?"non-zero":"zero"), altstr, actualval); 
      perror("err code: ");
      abort();
      return actualval;
    }
  })
  #define __tic_errcheckZ(fn) __tic_errcheck_helper(fn, 0, 0, #fn)
  #define __tic_errcheckZopt(fn,opt) __tic_errcheck_helper(fn, 0, opt, #fn)
  #define __tic_errcheckNZ(fn) __tic_errcheck_helper(fn, 1, 1, #fn)
#else /* compile away */
  #define __tic_errcheckZ(fn)  (fn)
  #define __tic_errcheckZopt(fn,opt)  (fn)
  #define __tic_errcheckNZ(fn) (fn)
#endif

#if defined (THREAD)  /* Solaris Threads */

#define __tic_lock_t mutex_t
#define __tic_cond_t cond_t

/* If at runtime we know there's only one processor on this box, we 
   don't need to perform synchronization. */
#define __tic_lock_init(lock)      \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(mutex_init(lock, USYNC_THREAD, NULL)) : 0)

INLINE_MODIFIER __tic_lock_t __tic_new_lock(void) 
INLINE({                                  
  __tic_lock_t var = DEFAULT_MUTEX;
  return var;                                    
})

#define __tic_lock_initializer  __tic_new_lock()
#define __tic_lock_decl_initializer DEFAULT_MUTEX

#define __tic_cond_init(cond, lock)   \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(cond_init(cond, USYNC_THREAD, NULL)) : 0)

INLINE_MODIFIER __tic_cond_t __tic_new_cond(void) 
INLINE({                                  
  __tic_cond_t var = DEFAULTCV;
  return var;                                   
})

#define __tic_cond_initializer __tic_new_cond()
#define __tic_cond_decl_initializer DEFAULTCV


#define __tic_lock(lock)           \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(mutex_lock(lock)) : 0)
#define __tic_try_lock(lock)       \
  ((MYBOXPROCS > 1) ? __tic_errcheckZopt(mutex_trylock(lock), EBUSY) : 1)
#define __tic_unlock(lock)         \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(mutex_unlock(lock)) : 0)
#define __tic_mon_wait(mon)            \
  do { if (MYBOXPROCS > 1) __tic_errcheckZ(cond_wait(&(mon->condvar), &(mon->lock))); } while(0)
#define __tic_mon_timedwait(mon, tm)   do {					\
	if (MYBOXPROCS > 1) \
        __tic_errcheckZopt(cond_timedwait(&(mon->condvar), &(mon->lock), tm), ETIMEDOUT);	\
	else ___tic_sleepuntil(tm);						\
	} while(0)
#define __tic_mon_signal(mon)          \
  do { if (MYBOXPROCS > 1) __tic_errcheckZ(cond_signal(&(mon->condvar))); } while(0)
#define __tic_mon_broadcast(mon)       \
  do { if (MYBOXPROCS > 1) __tic_errcheckZ(cond_broadcast(&(mon->condvar))); } while(0)

#define __tic_lock_destroy(lock)   \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(mutex_destroy(lock)) : 0)
#define __tic_cond_destroy(cond)   \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(cond_destroy(cond)) : 0)

#define ___tic_time_t	timestruc_t

#elif defined(PTHREAD)

#define __tic_lock_t pthread_mutex_t
#define __tic_cond_t pthread_cond_t

/* If at runtime we know there's only one processor on this box, we 
   don't need to perform synchronization. */
#if EXTRA_SYNC_CHECKING && defined(LINUX_THREADS)
INLINE_MODIFIER __tic_lock_t __tic_new_lock(void) 
INLINE({      
  __tic_lock_t lock;
  pthread_mutexattr_t attr; 
  __tic_errcheckZ(pthread_mutexattr_init(&attr));
  __tic_errcheckZ(pthread_mutexattr_setkind_np(&attr, PTHREAD_MUTEX_ERRORCHECK_NP));
  __tic_errcheckZ(pthread_mutex_init(&lock, &attr));
  __tic_errcheckZ(pthread_mutexattr_destroy(&attr));
  return lock;
  })
#define __tic_lock_init(lock)   \
   ((MYBOXPROCS > 1) ? *lock = __tic_new_lock(), 0 : 0)
#else
#define __tic_lock_init(lock)      \
   ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_mutex_init(lock, NULL)) : 0)
INLINE_MODIFIER __tic_lock_t __tic_new_lock(void) 
INLINE({      
  __tic_lock_t var = PTHREAD_MUTEX_INITIALIZER;  
  return var;                                    
})
#endif

#define __tic_lock_decl_initializer  PTHREAD_MUTEX_INITIALIZER
#define __tic_lock_initializer  __tic_new_lock()

#define __tic_cond_init(cond, lock)   \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_cond_init(cond, NULL)) : 0)

INLINE_MODIFIER __tic_cond_t __tic_new_cond(void) 
INLINE({                                  
  __tic_cond_t var = PTHREAD_COND_INITIALIZER;  
  return var;                                   
})

#define __tic_cond_initializer __tic_new_cond()
#define __tic_cond_decl_initializer PTHREAD_COND_INITIALIZER

#define __tic_lock(lock)           \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_mutex_lock(lock)) : 0)
#define __tic_try_lock(lock)       \
  ((MYBOXPROCS > 1) ? __tic_errcheckZopt(pthread_mutex_trylock(lock), EBUSY) : 1)
#define __tic_unlock(lock)         \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_mutex_unlock(lock)) : 0)

#ifdef CAUTIOUS_MONITORS
/* some pthread implementations need special help for wait/notify -
   specifically, IRIX pthreads frequently return prematurely from a wait
      so we must explicitly check the condition upon return 
   the extra fields in the monitor data structure are protected by the lock,
   which already must be held in order to call these wait/notify functions */
#define __tic_mon_wait(mon)   do { if (MYBOXPROCS > 1) {   		\
  mon->numwaiting++; /* about to wait */				\
  do {									\
   __tic_errcheckZ(pthread_cond_wait(&(mon->condvar), &(mon->lock)));   \
   } while (!mon->numsignaled);						\
  mon->numwaiting--; /* we're no longer waiting */			\
  mon->numsignaled--; /* we're awaking - consume one signal */ 		\
  } } while(0)
#define __tic_mon_timedwait(mon, tm)   do { if (MYBOXPROCS > 1) {   	\
  int retval = 0;							\
  mon->numwaiting++; /* about to wait */				\
  do {									\
   retval = __tic_errcheckZopt(pthread_cond_timedwait(&(mon->condvar), &(mon->lock), tm), ETIMEDOUT);	\
   } while (!mon->numsignaled && retval != ETIMEDOUT);			\
  mon->numwaiting--; /* we're no longer waiting */			\
  if (retval != ETIMEDOUT) mon->numsignaled--; /* we're awaking - consume one signal */ \
  } else ___tic_sleepuntil(tm); } while(0)
#define __tic_mon_signal(mon)   do { if (MYBOXPROCS > 1) {  		\
  if (mon->numwaiting > mon->numsignaled) /* some waiter still needs a signal */ \
    mon->numsignaled++; /* add one signal */				\
  __tic_errcheckZ(pthread_cond_signal(&(mon->condvar)));		\
  } } while(0)
#define __tic_mon_broadcast(mon)   do { if (MYBOXPROCS > 1) {  		\
  mon->numsignaled = mon->numwaiting; /* ensure a signal for each current waiter */ \
  __tic_errcheckZ(pthread_cond_broadcast(&(mon->condvar)));		\
  } } while(0)
#else
#define __tic_mon_wait(mon)   \
   do { if (MYBOXPROCS > 1) __tic_errcheckZ(pthread_cond_wait(&(mon->condvar), &(mon->lock))); } while (0)
#define __tic_mon_timedwait(mon, tm)   do {						\
    if (MYBOXPROCS > 1) __tic_errcheckZopt(pthread_cond_timedwait(&(mon->condvar), &(mon->lock), tm),ETIMEDOUT);	\
    else  ___tic_sleepuntil(tm);							\
    } while (0)
#define __tic_mon_signal(mon)    \
   do { if (MYBOXPROCS > 1) __tic_errcheckZ(pthread_cond_signal(&(mon->condvar))); } while (0)
#define __tic_mon_broadcast(mon) \
   do { if (MYBOXPROCS > 1) __tic_errcheckZ(pthread_cond_broadcast(&(mon->condvar))); } while (0)
#endif

#define __tic_lock_destroy(lock)   \
   ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_mutex_destroy(lock)) : 0)
#define __tic_cond_destroy(cond)   \
   ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_cond_destroy(cond)) : 0)

#define ___tic_time_t	struct timespec

#endif

#endif /* NO_SMP For machines like the NOW */  

/* Timer-related macros/functions */
/* ___tic_time_t is defined separately for each platform above */

/* set a time structure for given millis from right now */
#define ___tic_set_wait_timer(tm, millis) do { 				\
    struct timeval now;							\
    gettimeofday(&now, NULL);						\
    (tm)->tv_sec = now.tv_sec;						\
    (tm)->tv_sec += (now.tv_usec + millis*1000)/1000000;		\
    (tm)->tv_nsec = ((now.tv_usec + millis*1000)%1000000) * 1000;	\
    } while (0)

/* block until the given absolute time has approximately been reached */
#define ___tic_sleepuntil(tm) do {				\
	int millis;						\
	struct timeval now;					\
	gettimeofday(&now, NULL);				\
	millis = ((tm)->tv_sec - now.tv_sec) * 1000 +		\
	         ((tm)->tv_nsec / 1000 - now.tv_usec) / 1000;	\
	if (millis > 0) m5sleepJmT6Thread4lang4java(millis);	\
	} while (0)

/* return the number of milliseconds from now until the time indicated by tm
 * (will be negative if the time indicated by tm has already passed)
 */
INLINE_MODIFIER int ___tic_millisuntil(___tic_time_t* tm)
INLINE ({
    int millis;                                             
    struct timeval now;                                     
    gettimeofday(&now, NULL);                               
    millis = ((tm)->tv_sec - now.tv_sec) * 1000 +           
             ((tm)->tv_nsec / 1000 - now.tv_usec) / 1000;        
    return millis;
  })

#if HAVE_SCHED_YIELD
 #include <sched.h>
 #define tic_sched_yield() sched_yield()
#else
 extern void __os_yield(long usecs);
 #define tic_sched_yield() __os_yield(0)
#endif

#if HAVE_ALLOCA_H && ALLOCA_IN_C
 #include <alloca.h>
 #define tic_alloca(sz) alloca(sz)
 #define TRUE_ALLOCA 1
#else
 #define tic_alloca(sz) ti_malloc(sz)
 #define TRUE_ALLOCA 0
#endif

#if defined(GLUNIX)
#include <glib.h>
#endif

#include <unistd.h>
#include <comm_utils.h>

extern __tic_thread_key_t tic_thread_key;

INLINE_MODIFIER processInfo *myProcess() __attribute__ ((const));
INLINE_MODIFIER processInfo *myProcess()
INLINE({
  processInfo *pi; 
  __tic_thread_get_key_value(tic_thread_key, &pi);
  return (pi);
})


INLINE_MODIFIER Process myProcessNumber() __attribute__ ((const));
INLINE_MODIFIER Process myProcessNumber() 
INLINE({
  processInfo *pi; 
  __tic_thread_get_key_value(tic_thread_key, &pi);
  return (pi)->processNumber;
})

INLINE_MODIFIER Process myBoxProcessNumber() __attribute__ ((const));
INLINE_MODIFIER Process myBoxProcessNumber() 
INLINE({
  processInfo *pi; 
  __tic_thread_get_key_value(tic_thread_key, &pi);
  return (pi)->boxProcessNumber;
})

EXTERN __tic_lock_t AM_lock;

/* ........................
 *
 * Message Layer
 *
 * ........................
 */

#define LOWWORD(i)    ((julong)(i) & 0xFFFFFFFFUL)
#define HIGHWORD(i)   LOWWORD((julong)(i) >> 32) 

#define SETLOWWORD(ptr, i)  (*(ptr)) = (HIGHWORD(*(ptr)) << 32) | LOWWORD(i)
#define SETHIGHWORD(ptr, i) (*(ptr)) = (LOWWORD(((julong)(i))) << 32) | LOWWORD(*(ptr))


#ifdef COMM_AM2
#if defined(COMM_AMUDP)
#include <amudp.h>
#include <amudp_spmd.h>
#define AM2_HAS_HUGE_SEGMENTS 0 /* supports them, but no perf advantage */ 
#elif defined(COMM_AMMPI)
#include <ammpi.h>
#include <ammpi_spmd.h>
#define AM2_HAS_HUGE_SEGMENTS 0 /* supports them, but no perf advantage */ 
#elif defined(COMM_AMLAPI)
#include <lapi.h>
#include <amlapi.h>
#define AM2_HAS_HUGE_SEGMENTS 1 
#elif defined(COMM_AMNOW)
#define inline INLINE_MODIFIER_FOR_SOLARIS_AM
#include <am.h>
#define AM2_HAS_HUGE_SEGMENTS 0 /* OS limit for DMA registration prevents this */ 
#undef inline
#elif defined(COMM_AMVIA)
#include <amvia.h>
#define AM2_HAS_HUGE_SEGMENTS 0 /* OS limit for DMA registration prevents this */ 
#else
#error unknown AM implementation
#endif

#include <tic_am2_macros.h>
 
#define TIC_TAG		18

extern eb_t __sc_bundle;
extern ep_t __sc_endpoint;
#define ADVANCE_SEND_ENDPOINT
#define CUR_ENDPOINT          __sc_endpoint

#else /* AMII */
#define tic_poll()
#endif

/* ........................
 *
 * Synchronization counters
 *
 * ........................
 */
typedef volatile int Counter;


#ifdef COMM_AM2
INLINE_MODIFIER void sync_ctr(Counter *ctr)
INLINE({
	while(*ctr) {
	  tic_poll();
	}
})
#else
#define sync_ctr(a) 
#endif /* AMII */

#ifdef COMM_AM2
INLINE_MODIFIER int is_sync_ctr(Counter *ctr)
INLINE({
  tic_poll(); 
  return !*ctr;
})
#else
#define is_sync_ctr(a)  (1)
#endif /* AMII */


extern int politep;


/* ...............
 *
 * Global pointers
 *
 * ...............
 */

#include <pointers.h>


/* .............
 *
 * Handlers
 *
 * .............
 */

#include <handlers.h>

/* .............
 *
 * Global memory
 *
 * .............
 */

/* Constants specifying when to use segments
 * (these should go somewhere else... -TvE */
#define		STORE_STRUCT_LEN	60
#define		PUT_STRUCT_LEN		32

/* for distributed GC performance, it helps to know whether bulk transfers may contain pointers */
typedef enum {
  tic_no_ptrs, /* no embedded pointers at all */
  tic_gp_only, /* the only embedded pointers are global pointers */
  tic_has_ptrs /* may have global or local pointers embedded */
} tic_ptr_embedding;

/* Ack-ed reads and writes */
#include <mem.h>

/* Non-ack-ed stores */
#include <store.h>

#include <bulk.h>

/*************************************************************
 *
 * Global communication
 * 
 *************************************************************/
#include <com.h>

extern int usingAMSPMD; /* true for workers using the AMUDP/AMMPI SPMD job startup API */

/* Enrionment variable handling */

/* getenvMaster() retrieves environment information from the console node */
#ifdef COMM_AMUDP
  INLINE_MODIFIER const char *getenvMaster(const char *name)
  INLINE({
    if (usingAMSPMD) return AMUDP_SPMDgetenvMaster(name);
    else return getenv(name);
  })
#else
  INLINE_MODIFIER const char *getenvMaster(const char *name)
  INLINE({
    return getenv(name);
  })
#endif

#if !defined(HAVE_SETENV) && defined(HAVE_PUTENV)
  INLINE_MODIFIER int setenv(const char *name, const char *value, int overwrite)
  INLINE({
    if (overwrite || !getenv(name)) {
      char tmp[1024];
      assert(strlen(name)+strlen(value)+1 < 1024);
      sprintf(tmp,"%s=%s",name,value);
      return putenv(tmp);
    } else return 0;
  })
#endif

/* ..............
 *
 * Global control
 *
 * ..............
 */

#include <control.h> /* public stuff. Maybe should be here verbatim? */

/* ........................
 *
 * Barrier
 *
 * ........................
 */
#include <barrier.h>

/* ........................
 *
 * Monitors
 *
 * ........................
 */

#include <monitor.h>


/* .........................
 *
 * Timer library
 *
 * .........................
 */

EXTERN double get_seconds();


#endif
